import xml.etree.ElementTree as et


#
# ## Znode (to create serializable structure while keeping sanity) ##
#


class Znode(object):
    """Create serializable material structure, while keeping your sanity"""
    
    def __init__(self):
        self.graph = []
        self.pp = self.graph
        self.lp = self.graph
        self.cp = self.graph
    
    # [ {"node_name": [parent_pointer, attrib{}, children[]]} ]
    
    def append(self, name="", attr=[]) -> None:
        """Append nodes to the current level (parent)"""
        # append a new node to the current level
        ptr = self.pp
        self.lp.append({name: [ptr, attr, []]})
        # move cp to point at the data block of the last appended node
        #          get   |  current last  | name | datablock is always third (2)
        self.cp = self.lp[len(self.lp) - 1][name][2]
    
    def yank(self, graph=[]) -> None:
        """Yank an entire tree of nodes as sub-nodes to the last added node"""
        # current level pointer
        # ptr = self.pp
        self.cp.extend(graph)
        # WARNING! cp isn't touched!
        # self.cp = self.lp[len(self.lp) - 1][name][2] # this doesn't work (duh)
    
    def add(self, name="", attr=[]) -> None:
        """Add a sub-node to the last added node"""
        # add the new node to the last added's datablock
        ptr = self.lp
        self.cp.append({name: [ptr, attr, []]})
        # set pp to be on the previous (parent) level
        self.pp = self.lp
        # set lp to be on the level we just added to
        self.lp = self.cp
        # same as before
        # set cp to point at the last added's datablock.
        self.cp = self.lp[len(self.lp) - 1][name][2]
    
    def pointer_up(self) -> None:
        """Move the current level pointer up."""
        # not necessarily the best idea
        self.cp = self.lp
        # sounds a tad better
        self.lp = self.pp
        # get the name of the (only one) node in the last dict we know about.
        # [* iterable] ==> unpack
        name = [*self.lp[len(self.lp) - 1]][0]
        # set pp (parent) to be the same as the parent pointer in the data
        self.pp = self.lp[len(self.lp) - 1][name][0]
        # jump up a level
    
    def pointer_reset(self) -> None:
        """Move the current level pointer to root"""
        self.pp = self.graph
        self.lp = self.graph
        self.cp = self.graph


#
# ## Serializers (to write uniform-looking and dom-compliant files) ##
#


class XMLserializer(object):
    """Class for serializing XML"""
    
    def __init__(self, ind="\t"):
        self.ind = ind
    
    def write_file(self,
                   filepath,
                   graph=[],
                   ) -> None:
        """
        Basic write function. Getting them xmls written.
        """
        
        def indent(elem, level=0):
            ind = self.ind
            i = "\n" + level * ind
            if len(elem):
                if not elem.text or not elem.text.strip():
                    elem.text = i + ind
                if not elem.tail or not elem.tail.strip():
                    elem.tail = i
                for element in elem:
                    indent(element, level + 1)
                if not elem.tail or not elem.tail.strip():
                    elem.tail = i
            else:
                if level and (not elem.tail or not elem.tail.strip()):
                    elem.tail = i
        
        # seri
        
        def serialize(root, graph):
            for node in graph:
                
                # """ {"nodes": [ptr, {}, [{"name": [{}, []]}]]} """
                
                # unpack
                name = [*node][0]
                ptr, attr, cgraph = node[name]
                
                # """ nodes """
                # """ [ptr, {}, [{"name": [{}, []]}]] """
                
                for at in attr:
                    if type(attr[at]) is float:
                        attr[at] = f"{attr[at]:.6f}"
                    if type(attr[at]) is not str:
                        attr[at] = str(attr[at])
                # """ [{"name": [{}, []]}] """
                
                croot = et.SubElement(root, name, attr)
                if len(cgraph):
                    serialize(croot, cgraph)
        
        # TODO:
        # Initialize string to write with header
        
        # """ {"scene": [ptr, {attr}, []]} """
        
        root_name = root_data = None
        for root_name in graph[0]:
            root_data = graph[0][root_name]
            break
        
        root = et.Element(root_name, root_data[1])
        tree = et.ElementTree(root)
        
        serialize(root, root_data[2])
        
        indent(root)
        
        # And finally!
        with open(filepath, mode="w", encoding="UTF-8"):
            # "ugly" printing of xml, sadly "pretty" is hard.
            tree.write(filepath, encoding="UTF-8")


class JSXserializer(object):
    """Class for serializing JSX"""
    
    def __init__(self, ind="  "):
        self.ind = ind
    
    def write_file(self,
                   filepath,
                   graph=[],
                   ) -> None:
        """
        Basic write function. Getting them jsx written.
        """
        
        def ex_dict(d: dict, level) -> str:
            seri = ""
            ind = str(self.ind * level)
            
            def pr(dt) -> str:
                tmp = ""
                c = 0
                for key in dt:
                    val = dt[key]
                    
                    if type(val) is str:
                        tmp += f"{key}: '{val}'"
                    elif type(val) in {float, int}:
                        tmp += f"{key}: {val}"
                    elif type(val) is bool:
                        if val:
                            tmp += f"{key}: true"
                        else:
                            tmp += f"{key}: false"
                    elif type(val) is list:
                        tmp += f"{key}: {val}"
                    elif type(val) is dict:
                        if key == "m":
                            tmp2 = pr(val)
                            tmp += " m => {" + tmp2 + " }"
                        elif "m." in key:
                            tmp2 = pr(val)
                            tmp += f"{key}" + "({" + tmp2 + "})"
                        else:
                            tmp2 = pr(val)
                            tmp += "{{ " + tmp2 + " }}"
                    else:
                        tmp += f"{key}: {val}"
                    
                    c = c + 1
                    if len(dt.keys()) > c:
                        tmp += ", "
                    else:
                        tmp += " "
                
                return tmp
            
            i = 0
            for key in d:
                val = d[key]
                
                seri += "\n"
                
                tmp = pr(val)
                
                if key == "...":
                    seri += f"{ind}{self.ind}" + "{...{ " + tmp + "}}"
                elif key == "done":
                    seri += f"{ind}{self.ind}{key}=" + "{" + tmp + "}"
                else:
                    seri += f"{ind}{self.ind}{key}=" + "{{ " + tmp + "}}"
                
                i = i + 1
                # if len(d.keys()) > i:
                #     seri += ","
            
            return seri
        
        # seri
        
        def serialize(graph, level=0) -> str:
            seri = ""
            for node in graph:
                # unpack
                name = [*node][0]
                # get datablock
                ptr, attr, cgraph = node[name]
                # calc indent
                ind = str(self.ind * level)
                # begin tag
                seri += f"{ind}<{name}"
                
                if len(attr) and type(attr) is dict:
                    seri += ex_dict(attr, level)
                    seri += ">\n"
                else:
                    seri += ">\n"
                # move on to serialize children (if any)
                if len(cgraph):
                    seri += (serialize(cgraph, level+1))
                # close tag
                seri += f"{ind}</{name}>\n"
            
            return seri
        
        seri_string = ""
        seri_string += serialize(graph)
        
        # And finally!
        with open(filepath, mode="w", encoding="UTF-8") as fw:
            fw.write(seri_string)
