import os
from os import path, listdir
import platform
import gc
import subprocess
import shutil
import json
# from mathutils import Matrix, Vector, Color
# from bpy_extras import io_utils, node_shader_utils

import bpy
from bpy_extras.wm_utils.progress_report import (ProgressReport, ProgressReportSubstep)

# from ..utils_ogre import ogre_types, ogre_export

from ..utils_general import znode, zutil
from .mw_switch import sw_object
from .types.mw_meta import IndexJSX, WhereJSON, WebtablesJSON
from .types.mw_mesh import Mesh
from .types.mw_material import Material, Skybox

#
# ## Save ##
#


def save(context,
         sr,
         filepath,
         *,
         export_selected,
         export_hidden,
         export_armature,
         export_animation,
         export_physics,
         ex_phys_auto_props,
         ex_phys_bias,
         export_materials,
         do_binary,
         do_clean,
         do_unique_names,
         apply_modifiers,
         phy_mesh_quality,
         global_matrix,
         path_exec_root
         ) -> set:
    """
    The main function called from __init__.py to save the scene
    """
    
    import cProfile, pstats, io
    from pstats import SortKey
    pr = cProfile.Profile()
    pr.enable()
    
    with ProgressReport(context.window_manager) as progress:
        progress.enter_substeps(1)
        
        write_loop(context, sr, filepath, path_exec_root,
        EX_SELECTED=export_selected,
        EX_HIDDEN=export_hidden,
        EX_ARMATURE=export_armature,
        EX_ANIMATION=export_animation,
        EX_PHYSICS=export_physics,
        EX_PHYS_AUTO_PROPS=ex_phys_auto_props,
        EX_PHYS_BIAS=ex_phys_bias,
        EX_MATERIALS=export_materials,
        EX_BINARY=do_binary,
        EX_UNIQUE_NAMES=do_unique_names,
        EX_CLEAN=do_clean,
        EX_APPLY_MODIFIERS=apply_modifiers,
        EX_PHY_MESH_QUALITY=phy_mesh_quality,
        EX_GLOBAL_MATRIX=global_matrix,
        progress=progress
        )
        
        progress.leave_substeps()
    
    # print("B2MW finished: %.4f sec" % (time.time() - time_start))
    
    pr.disable()
    s = io.StringIO()
    ps = pstats.Stats(pr, stream=s).sort_stats(SortKey.TIME)
    ps.print_stats(20)
    ps = pstats.Stats(pr, stream=s).sort_stats(SortKey.CUMULATIVE)
    ps.print_stats(20)
    print(s.getvalue())
    
    # Return finished state; if we had errors, it would yield a traceback anyway.
    return {'FINISHED'}


#
# ## Write call ##
#


def write_loop(context, sr, filepath, path_exec_root,
               EX_SELECTED,
               EX_HIDDEN,
               EX_ARMATURE,
               EX_ANIMATION,
               EX_PHYSICS,
               EX_MATERIALS,
               EX_PHYS_AUTO_PROPS,
               EX_PHYS_BIAS,
               EX_BINARY,
               EX_UNIQUE_NAMES,
               EX_CLEAN,
               EX_APPLY_MODIFIERS,
               EX_PHY_MESH_QUALITY,
               EX_GLOBAL_MATRIX,
               progress: ProgressReport
               ) -> None:
    
    # WRITE LOOP BEGIN
    rtv = None
    import time
    time_start = time.time()
    
    with ProgressReportSubstep(progress, 2, "Exporting MaxWhere scene...", "Export finished!") as subprogress1:
        
        # construct paths
        path_main, ext = path.splitext(filepath)
        path_dir_misc = path.join(path_main, "misc")
        path_dir_res = path.join(path_main, "resources")
        path_dir_thumb = path.join(path_main, "thumbs")
        
        path_converter = path_dir_res
        # are we running on windows?
        if os.name == "nt":
            path_converter = path.join("C:\\", "ProgramData", "MWexporter_TEMP")
            # ensure directory exists
            if not os.path.exists(path_converter):
                os.makedirs(path_converter)
        
        if path.exists(path_main):
            # print(f"Warning! the path: \"{path_main}\" exists already. Deleting.")
            shutil.rmtree(path_main)
        
        paths = list()
        paths.append(path_main)
        paths.append(path_dir_misc)
        paths.append(path_dir_res)
        paths.append(path_dir_thumb)
        
        for p in paths:
            os.makedirs(p)
        
        # stopper
        if not os.path.exists(path_main):
            sr.report({"ERROR"}, "Something went horribly wrong. Please make sure the export directory exists.")
            return None
        
        # ## Set rotations to be quats ##
        # cause otherwise things go wrong.
        
        for obj in bpy.data.objects:
            obj.rotation_mode = "QUATERNION"
        # just to make absolutely sure.
        context.view_layer.update()
        
        # ## Lists ##
        
        list_mesh = []
        list_material = []
        list_light = []
        list_camera = []
        map_webtable = {}
        
        # struct ixtr
        # {
        #  ixtr* parent;
        #  IGameNode* node;
        #  int type;
        #  bool unique;
        #  std::string name;
        #  std::vector<ixtr> children;
        # };
        
        # ## Object switch + index ##
        
        index_jsx = IndexJSX()
        index_jsx.init_root()
        
        rtv = sw_object(context, sr,
                        EX_SELECTED, EX_HIDDEN, EX_ARMATURE, EX_ANIMATION, EX_PHYSICS, EX_MATERIALS, EX_GLOBAL_MATRIX,
                        list_mesh, list_material, list_light, list_camera, map_webtable, index_jsx)
        if rtv is None:
            return None
        
        # ## Export where.json ##
        
        def ex_wherejson():
            where_json = WhereJSON()
            scene = context.scene
            
            p_name = ""
            p_creator = ""
            
            if "mxw_props" in scene.keys():
                if "mxw_name" in scene.keys():
                    p_name = scene["mxw_name"]
                if "mxw_creator" in scene.keys():
                    p_creator = scene["mxw_creator"]
            
            tmp_dict = dict()
            # tmp_dict[""] = data
            
            tmp_dict["deploy-name"]           = p_name
            tmp_dict["meta"] = dict()
            tmp_dict["meta"]["creator"]       = p_creator
            tmp_dict["name"]                  = p_name
            
            zutil.dict_merge(where_json.data, tmp_dict)
            
            where_json.pull_scene_data(context.scene, list_camera, EX_GLOBAL_MATRIX)
            
            path_where_json = path.join(path_main, "where.json")
            with open(path_where_json, mode="w") as f:
                f.write(json.dumps(where_json.data, indent="  "))
            
            return True
        
        # print(f"Exporting where.json", flush=True)
        rtv = ex_wherejson()
        if rtv is None:
            return None
        
        # ## Export webtables.json ##
        
        def ex_webtablesjson():
            webtables_json = WebtablesJSON()
            
            for ixi in sorted(map_webtable.keys()): # very neat, very orderly
                obj = map_webtable[ixi]
                webtables_json.pull_table_data(obj, ixi, EX_GLOBAL_MATRIX)
            
            path_webtables_json = path.join(path_main, "webtables.json")
            with open(path_webtables_json, mode="w") as f:
                f.write(json.dumps(webtables_json.data, indent="  "))
            
            return True
        
        # print(f"Exporting webtables.json", flush=True)
        rtv = ex_webtablesjson()
        if rtv is None:
            return None
        
        # ## Export index.jsx ##
        
        def ex_indexjsx():
            # serialize
            seri_jsx = znode.JSXserializer()
            
            path_index_jsx = path.join(path_main, "index.jsx")
            graph = index_jsx.graph
            seri_jsx.write_file(path_index_jsx, graph)
            
            return True
        
        # print(f"Exporting index.jsx", flush=True)
        rtv = ex_indexjsx()
        if rtv is None:
            return None
        
        # free memory
        gc.collect()
        
        # ## Export XML, Binary & Phy meshes ##
        
        def xml_convert(name_mesh):
            print(f"INFO: Calling XML converter on {name_mesh}:")
            cmd = [
                path.join(path_converter, "OgreXMLConverter.exe"),
                f"{path.join(path_converter, name_mesh)}.mesh.xml",
                f"{path.join(path_converter, name_mesh)}.mesh"
                ]
            proc = subprocess.Popen(cmd)
            
            while proc.poll() is None:
                time.sleep(0.200)
        
        def gen_physmesh(name_mesh, num_verts, try_num):
            if EX_PHY_MESH_QUALITY == 0:
                meshp_args = ""
            elif EX_PHY_MESH_QUALITY == -1:
                meshp_args = "-meshDecimation 10 10 0.01"
            elif EX_PHY_MESH_QUALITY == -2:
                meshp_args = "-meshDecimation 10 10 0.1"
            elif EX_PHY_MESH_QUALITY == -3:
                meshp_args = "-meshDecimation 10 10 0.2"
            elif EX_PHY_MESH_QUALITY == -4:
                meshp_args = "-convexDecomposition 8000 32"
            elif EX_PHY_MESH_QUALITY == -5:
                meshp_args = "-convexHull"
            else:
                meshp_args = "-meshDecimation 10 10 0.01"
            
            path_mesh = path.join(path_converter, f"{name_mesh}.mesh.xml")
            path_mesh_obj = path.join(path_converter, f"{name_mesh}_phy.obj")
            
            print(f"INFO: Calling meshprocessor on {name_mesh}:")
            ratio = zutil.clamp((zutil.squish(num_verts / 200, EX_PHYS_BIAS, 8) - try_num * 0.01), 0.0, 0.40)
            arg = f"-meshDecimation 10 10 {ratio}"
            
            if not EX_PHYS_AUTO_PROPS:
                arg = meshp_args
            
            cmd = [path.join(path_exec_root, "ogrexmltools", "meshprocessor.exe"), "-import", path_mesh] + arg.split() + ["-saveCollisionShapes", path_mesh_obj]
            proc = subprocess.Popen(cmd)
            
            while proc.poll() is None:
                time.sleep(0.200)
        
        def ex_mesh(subp: ProgressReportSubstep, EX_BINARY, EX_PHYSICS):
            subp.enter_substeps(len(list_mesh) * 3, "Exporting meshes...")
            if EX_BINARY:
                path_ogre_main = path.join(path_exec_root, "ogrexmltools", "OgreMain.dll")
                path_ogre_xml = path.join(path_exec_root, "ogrexmltools", "OgreXMLConverter.exe")
                zutil.cpfile(path_ogre_main, path_converter)
                zutil.cpfile(path_ogre_xml, path_converter)
            if EX_PHYSICS:
                path_mp = path.join(path_exec_root, "ogrexmltools", "meshprocessor.exe")
                zutil.cpfile(path_mp, path_converter)
            
            for obj, mtx in list_mesh:
                mesh = Mesh()
                num_verts = mesh.serialize(obj, mtx, context.evaluated_depsgraph_get(), EX_GLOBAL_MATRIX, EX_ARMATURE, EX_ANIMATION, EX_APPLY_MODIFIERS)
                
                # IMPORTANT: sanitize name
                name_mesh = path.normpath(zutil.sanitize(mesh.name))
                path_mesh = path.join(path_dir_res, name_mesh)
                print(f"INFO: Done deduping {name_mesh}, final vertex count: , writing to xml file...")
                with open(f"{path_mesh}.mesh.xml", mode="w", encoding="UTF-8") as fp:
                    fp.write(mesh.sb)
                
                # free memory
                del mesh
                gc.collect()
                
                subp.step()
                
                # path_ogre = path.join(path_exec_root, "ogrexmltools")
                path_mesh_bin = path.join(path_converter, f"{name_mesh}.mesh")
                path_mesh_obj = path.join(path_converter, f"{name_mesh}_phy.obj")
                
                if EX_BINARY:
                    if path_converter != path_dir_res:
                        zutil.cpfile(path.join(path_dir_res, f"{name_mesh}.mesh.xml"), path_converter)
                    xml_convert(name_mesh)
                    print(f"INFO: XML converter finished on mesh: {name_mesh}")
                    time.sleep(0.050)
                    if path.exists(path_mesh_bin):
                        if path_converter != path_dir_res:
                            print(f"INFO: copying mesh: {path_mesh_bin} to {path_dir_res}")
                            zutil.cpfile(path_mesh_bin, path_dir_res)
                            zutil.rmfile(path.join(path_converter, f"{name_mesh}.mesh"))
                            zutil.rmfile(path.join(path_converter, f"{name_mesh}.mesh.xml"))
                    else:
                        print(f"!! ERROR !!: XML converter failed to convert: {name_mesh}")
                        with open(path.join(path_dir_res, "failed_binary.txt"), mode="a", encoding="UTF-8") as fw:
                            fw.write(f"{name_mesh}\n")
                    subp.step()
                else:
                    subp.step()
                
                if EX_PHYSICS:
                    tries = 0
                    done = False
                    while tries < 3:
                        if path_converter != path_dir_res:
                            zutil.cpfile(path.join(path_dir_res, f"{name_mesh}.mesh.xml"), path_converter)
                        gen_physmesh(name_mesh, num_verts, tries)
                        print(f"INFO: meshprocessor finished on mesh: {name_mesh}")
                        time.sleep(0.050)
                        if path.exists(path_mesh_obj):
                            done = True
                            if path_converter != path_dir_res:
                                print(f"INFO: copying mesh: {path_mesh_obj} to {path_dir_res}")
                                zutil.cpfile(path_mesh_obj, path_dir_res)
                                zutil.rmfile(path_mesh_obj)
                                zutil.rmfile(path.join(path_converter, f"{name_mesh}.mesh.xml"))
                            break
                        else:
                            print(f"! WARN !: meshprocessor failed try #{tries + 1} on converting {name_mesh}")
                        tries = tries + 1
                    if not done:
                        print(f"!! ERROR !!: meshprocessor failed to convert: {name_mesh}")
                        with open(path.join(path_dir_res, "failed_physics.txt"), mode="a", encoding="UTF-8") as fw:
                            fw.write(f"{name_mesh}\n")
                    subp.step()
                else:
                    subp.step()
            
            if EX_BINARY:
                path_ogre_main = path.join(path_converter, "OgreMain.dll")
                path_ogre_xml = path.join(path_converter, "OgreXMLConverter.exe")
                zutil.rmfile(path_ogre_main)
                zutil.rmfile(path_ogre_xml)
            if EX_PHYSICS:
                path_mp = path.join(path_converter, "meshprocessor.exe")
                zutil.rmfile(path_mp)
            
            subp.leave_substeps("Mesh export done.")
            return True
        
        rtv = ex_mesh(subprogress1, EX_BINARY, EX_PHYSICS)
        if rtv is None:
            return None
        
        # ## Export materials ##
        
        def ex_mats(subp: ProgressReportSubstep, EX_UNIQUE_NAMES):
            # ## Construct material list ##
            path_reflection_map = None
            
            if "mxw_def_rfx" in context.scene.keys() and context.scene["mxw_def_rfx"] != "":
                path_reflection_map = context.scene["mxw_def_rfx"]
                src = path.join(bpy.path.abspath("//"), path_reflection_map)
                # if we can't find the file, we exit here.
                if not path.exists(src):
                    sr.report({"ERROR"}, f"File > {path_reflection_map} < not found for default reflection map. Make sure your MxW properties data is correct.")
                    return None
            else:
                path_reflection_map = "default.dds"
                src = path.join(path_exec_root, "res", "default.dds")
            
            # copy the reflection map over
            zutil.cpfile(src, path_dir_res)
            
            subp.enter_substeps(len(list_material), "Exporting materials...")
            
            for material in list_material:
                mat = Material()
                rtv = mat.pull(sr, context, material, path_dir_res, path_reflection_map, EX_UNIQUE_NAMES, EX_CLEAN)
                if rtv is None:
                    return None
                path_mat = path.join(path_dir_res, f"{material.name}.material")
                strb = mat.serialize()
                with open(path_mat, mode="w", encoding="UTF-8") as fw:
                    fw.write(strb)
                
                subp.step()
            
            # deal with the environment material
            skybox = Skybox()
            rtv = skybox.pull(sr, context, path_exec_root, path_dir_res, EX_CLEAN)
            if rtv is None:
                return None
            
            strb = skybox.serialize()
            path_env = path.join(path_dir_res, f"FXUnlitSkyboxLatlong.material")
            with open(path_env, mode="w", encoding="UTF-8") as fw:
                fw.write(strb)
            
            subp.leave_substeps("Material export done.")
            return True
        
        if EX_MATERIALS:
            # print(f"Exporting Material files", flush=True)
            rtv = ex_mats(subprogress1, EX_UNIQUE_NAMES)
            if rtv is None:
                return None
            # free memory
            gc.collect()
        else:
            subprogress1.step()
        
        # ## Clean export folder of unneded resources ##
        
        def ex_clean():
        
            def anyin(tu, st):
                for x in tu:
                    if x in st:
                        return True
                return False
            
            # init file list
            path_garbage = list()
            # fill list with files to convert
            for fp in listdir(path_dir_res):
                fp2 = path.join(path_dir_res, fp)
                gb_ext = ".mesh.xml"
                if path.isfile(fp2) and (gb_ext in fp):
                    path_garbage.append(fp2)
            
            # print(f"INFO - Garbage: {path_garbage}")
            
            for fp in path_garbage:
                zutil.rmfile(fp)
            
            return True
        
        if EX_CLEAN:
            # print(f"INFO - Cleaning export folder", flush=True)
            rtv = ex_clean()
            if rtv is None:
                return None
    
    if rtv is not None:
        sr.report({"INFO"}, f"MaxWhere export finished! Time taken: {(time.time() - time_start):.2f} s")
    
    #                   #
    #                   #
    # END OF WRITE LOOP #
    #                   #
    #                   #