import os
from os import path, listdir
import pathlib
import subprocess
import shutil
import re

# from mathutils import Matrix, Vector, Color
# from bpy_extras import io_utils, node_shader_utils

import bpy
from bpy_extras.wm_utils.progress_report import (ProgressReport, ProgressReportSubstep)

from ..utils_general import zutil
from ..utils_external import export_obj


#
# ## Save ##
#


def save(context,
         sr,
         filepath,
         *,
         export_selected,
         do_clean,
         global_matrix,
         path_exec_root
         ) -> set:
    """
    The main function called from __init__.py to save DXT maps
    """
    
    import cProfile, pstats, io
    from pstats import SortKey
    pr = cProfile.Profile()
    pr.enable()
    
    with ProgressReport(context.window_manager) as progress:
        progress.enter_substeps(1)
        
        write_loop(context, sr, filepath, path_exec_root,
        EX_SELECTED=export_selected,
        EX_CLEAN=do_clean,
        EX_GLOBAL_MATRIX=global_matrix,
        progress=progress
        )
        
        progress.leave_substeps()
    
    pr.disable()
    s = io.StringIO()
    sortby = SortKey.CUMULATIVE
    ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
    ps.print_stats(20)
    print(s.getvalue())
    
    # Return finished state; if we had errors, it would yield a traceback anyway.
    return {'FINISHED'}


#
# ## Write call ##
#


def write_loop(context, sr, filepath, path_exec_root,
               EX_SELECTED,
               EX_CLEAN,
               EX_GLOBAL_MATRIX,
               progress: ProgressReport
               ) -> None:
    
    # WRITE LOOP BEGIN
    rtv = None
    import time
    time_start = time.time()
    
    with ProgressReportSubstep(progress, 7, "Exporting Physical Objects...", "Export finished!") as subprogress1:
        
        # construct paths
        path_main, ext = path.splitext(filepath)
        path_main, ext = path.splitext(path_main)
        
        # if path.exists(path_main):
        #     # print(f"Warning! the path: \"{path_main}\" exists already. Deleting.")
        #     shutil.rmtree(path_main)
        #     pass
        
        paths = list()
        paths.append(path_main)
        
        for p in paths:
            os.makedirs(p, exist_ok=True)
        
        # stopper
        if not os.path.exists(path_main):
            sr.report({"ERROR"}, "Something went horribly wrong. Please make sure the export directory exists.")
            return None
        
        # ## Set rotations to be quats ##
        # cause otherwise things go wrong.
        
        for obj in bpy.data.objects:
            obj.rotation_mode = "QUATERNION"
        # just to make absolutely sure.
        context.view_layer.update()
        
        depsgraph = context.evaluated_depsgraph_get()
        
        objls = []
        for obj in bpy.data.objects:
            if obj.type in {"MESH", "SURFACE", "FONT", "LATTICE"} and obj.data:
                if (obj.select_get() is False) and EX_SELECTED:
                    # skip
                    # print(f"Skipping (not selected): {obj}")
                    continue
                
                objls.append((obj, obj.matrix_world))
                if obj.is_instancer:
                    for dup in context.evaluated_depsgraph_get().object_instances:
                        if dup.parent and dup.parent.original == obj:
                            objls.append((dup.instance_object.original, dup.matrix_world.copy()))
                
                
        
        
        for t_obj in objls:
            obj, mtx = t_obj
            
            path_mesh = path.join(path_main, obj.data.name + ".phy_tmp.obj")
            
            export_obj.write_file(path_mesh, [obj], depsgraph,
                                context.scene,
                                EXPORT_UV=False,
                                EXPORT_MTL=False,
                                EXPORT_BLEN_OBS=False,
                                EXPORT_GLOBAL_MATRIX=mtx,
                                progress=subprogress1
                                )
        
        
        for t_obj in objls:
            obj, mtx = t_obj
            
            path_mesh_orig = path.join(path_main, obj.data.name + ".phy_tmp.obj")
            path_mesh_tmp = path.join(path_exec_root, "ogrexmltools", obj.data.name + "_phy.obj")
            
            meshp_cmd = [path.join(path_exec_root, "ogrexmltools", "meshprocessor.exe"), "-import", path_mesh_orig, "-saveCollisionShapes", path_mesh_tmp]
            
            meshp_proc = subprocess.Popen(meshp_cmd)
            
            while meshp_proc.poll() is None:
                time.sleep(0.200)
            
            zutil.cpfile(path_mesh_tmp, path_main)
            zutil.rmfile(path_mesh_tmp)
            
        
        
        # ## Clean export folder of unneded resources ##
        
        def ex_clean():
            for t_obj in objls:
                obj, mtx = t_obj
                
                path_mesh_orig = path.join(path_main, obj.data.name + ".phy_tmp.obj")
                zutil.rmfile(path_mesh_orig)
            return True
        
        if EX_CLEAN:
            # print(f"INFO - Cleaning export folder", flush=True)
            rtv = ex_clean()
            if rtv is None:
                return None
            subprogress1.step("Export folder cleaned.")
        else:
            subprogress1.step()
        
    
    if rtv is not None:
        sr.report({"INFO"}, f"MaxWhere export finished! Time taken: {(time.time() - time_start):.2f} s")
