import bpy
from mathutils import Matrix, Vector, Quaternion
from os import path

from ...utils_general import znode, zutil


class IndexJSX(object):
    """docstring for IndexJSX."""
    
    # Constructor
    def __init__(self):
        self.graph = []
        self.pp = self.graph
        self.lp = self.graph
        self.cp = self.graph
    
    def init_root(self) -> None:
        self.graph = \
        [
            {
                "root": [self.graph, {}, []]
            }
        ]
        
        self.pp = self.graph
        self.lp = self.graph
        self.cp = self.graph[0]["root"][2]
    
    def append_node(self, obj: bpy.types.Object, obj_mtx: Matrix, EX_GLOBAL_MATRIX: Matrix, EX_PHYSICS) -> None:
        
        cp, graph = self.mknode(obj, obj_mtx, EX_GLOBAL_MATRIX, EX_PHYSICS)
        
        graph["node"][0] = self.pp
        self.lp.append(graph)
        self.cp = cp
    
    def add_node(self, obj: bpy.types.Object, obj_mtx: Matrix, EX_GLOBAL_MATRIX: Matrix, EX_PHYSICS) -> None:
        
        cp, graph = self.mknode(obj, obj_mtx, EX_GLOBAL_MATRIX, EX_PHYSICS)
        
        graph["node"][0] = self.lp
        self.cp.append(graph)
        
        self.pp = self.lp
        self.lp = self.cp
        self.cp = cp
    
    def mknode(self, obj: bpy.types.Object, obj_mtx: Matrix, EX_GLOBAL_MATRIX: Matrix, EX_PHYSICS) -> list():
        cp = None
        
        gm_trl, gm_rot, gm_scl = EX_GLOBAL_MATRIX.decompose()
        lm_trl, lm_rot, lm_scl = obj.matrix_local.decompose()
        
        pos = lm_trl * gm_scl
        # pos = pos @ trl
        quad = lm_rot
        # quad = obj.rotation_quaternion.copy()
        # quad.rotate(gm_rot)
        scale = lm_scl
        # scale = [lm_scl[0] * gm_scl[0], lm_scl[1] * gm_scl[1], lm_scl[2] * gm_scl[2]]
        # scale = [obj.scale[0] * scl[0], obj.scale[1] * scl[1], obj.scale[2] * scl[2]]
        
        # IMPORTANT: sanitize name
        name = path.normpath(zutil.sanitize(obj.name))
        if obj.data:
            name = path.normpath(zutil.sanitize(obj.data.name))
        
        # {name: [ptr, attr, []]}
        node = \
        {
            "node":
            [
                None,
                {
                    "position": {"x": pos[0], "y": pos[2], "z": -pos[1]},
                    "orientation": {"w": quad[0], "x": quad[1], "y": quad[3], "z": -quad[2]},
                    # "scale": {"x": scale[0], "y": scale[1], "z": scale[2]}
                },
                []
            ]
        }
        # cp = node["node"][2]
        
        if obj.type not in {"LIGHT"}:
            node["node"][1]["scale"] = {"x": scale[0], "y": scale[1], "z": scale[2]}
        
        if obj.type in {"MESH", "SURFACE", "FONT", "LATTICE"}:
            
            tmp = \
            {
                "mesh":
                [
                    node["node"][2],
                    {
                        "...": {"url": f"{name}.mesh"}
                    },
                    []
                ]
            }
            
            if EX_PHYSICS:
                tmp["mesh"][1]["physical"] = {"raycast": True, "shapes": [
                    {
                        "shape": "mesh",
                        "mass": 1,
                        "offset": { "x": 0, "y": 0, "z": 0 },
                        "rotation": { "w": 1, "x": 0, "y": 0, "z": 0 },
                        "mesh": f"{name}_phy.obj"
                    }
                ]}
                # tmp["mesh"][1]["done"] = {"m": {"m.addPhysicalShape": {"shape": "mesh", "mesh": f"{name}_phy.obj", "mass": 1}}}
            
            node["node"][2].append(tmp)
        
        elif obj.type in {"LIGHT"}:
            
            light = bpy.data.lights.get(obj.data.name)
            if light is None:
                raise BaseException(f"No light found with id {obj.data.name}")
            
            typ = "point"
            if light.type in {"SUN", "AREA"}:
                typ = "directional"
            elif light.type == "SPOT":
                typ = "spot"
            
            col = light.color
            power = light.energy
            
            if light.type == "SUN":
                power = zutil.scale(power, 0, 1, 0, 100)
            else:
                power = zutil.scale(power, 0, 1000, 0, 100)
            
            lim = 100
            if power > lim:
                power = lim
            
            c = []
            for v in col:
                c.append(v * power / lim)
            
            tmp = \
            {
                "light":
                [
                    node["node"][2],
                    {
                        "...": {"lighttype": typ},
                        "direction": {"x": 0, "y": -1, "z": 0}, # Ogre API light creation
                        "diffuse": {"r": c[0], "g": c[1], "b": c[2]},
                        "specular": {"r": c[0], "g": c[1], "b": c[2]}
                    },
                    []
                ]
            }
            
            node["node"][2].append(tmp)
        
        cp = node["node"][2]
        
        return [cp, node]
    
    def pointer_up(self) -> None:
        """Move the current level pointer up."""
        # not necessarily the best idea
        self.cp = self.lp
        # sounds a tad better
        self.lp = self.pp
        # get the name of the (only one) node in the last dict we know about.
        # [* iterable] ==> unpack
        name = [*self.lp[len(self.lp) - 1]][0]
        # set pp (parent) to be the same as the parent pointer in the data
        self.pp = self.lp[len(self.lp) - 1][name][0]
        # jump up a level
    
    def pointer_reset(self) -> None:
        """Move the current level pointer to root"""
        self.pp = self.graph
        self.lp = self.graph
        self.cp = self.graph


class WebtablesJSON(object):
    """docstring for WebtablesJSON"""
    
    def __init__(self):
        self.data = []
    
    def pull_table_data(self, table, ix, EX_GLOBAL_MATRIX):
        
        dt = dict()
        
        # px, py, pz = table.location @ EX_GLOBAL_MATRIX
        lm_trl, lm_rot, lm_scl = table.matrix_local.decompose()
        scl = EX_GLOBAL_MATRIX.decompose()[2]
        px, py, pz = lm_trl * scl
        
        q = lm_rot.copy()
        rot = Quaternion(Vector((0.707107, -0.707107, 0, 0)))
        q.rotate(rot)
        qw, qx, qy, qz = q
        
        size = [table.dimensions[0] * scl[0], table.dimensions[1] * scl[1], table.dimensions[2] * scl[2]]
        
        res_x = table["resolution_x"]
        res_y = table["resolution_y"]
        wh_ratio = size[0] / size[1]
        # hw_ratio = size[1] = size[0]
        
        if res_x == -1:
            res_x = 1600
        if res_y == -1:
            res_y = int(res_x / wh_ratio)
        
        dt["index"] = ix
        dt["url"] = ""
        dt["position"] = {
            "x": px,
            "y": pz,
            "z": -py
            }
        dt["orientation"] = {
            "x": qx,
            "y": qy,
            "z": qz,
            "w": qw
        }
        dt["size"] = {
            "width": size[0],
            "height": size[1]
        }
        dt["resolution"] = {
            "width": res_x,
            "height": res_y
        }
        dt["lastLoadedURL"] = "maxwhere:///default"
        dt["autoResolution"] = True
        
        self.data.append(dt)


class WhereJSON(object):
    """docstring for WhereJSON."""
    
    def __init__(self):
        self.data = {
            "deploy-name": None,
            "version": 0,
            "main": "index.jsx",
            "environment": {
                "background": {"r": 0.2, "g": 0.2, "b": 0.2, "a": 1},
                "ambient": {"r": 0.2, "g": 0.2, "b": 0.2, "a": 1},
                "skybox": {"material": "FXUnlitSkyboxLatlong"}},
            "camera": {
                "position": {"x": 0.0, "y": 0.0, "z": 0.0},
                "orientation": {"w": 1.0, "x": 0.0, "y": 0.0, "z": 0.0}},
            "local": "./resources",
            "meta": {"affiliation": None, "creator": None},
            "description": {"full": None, "short": None},
            "tags": [],
            "isPrivate": False,
            "isPublic": False,
            "mainTag": None,
            "name": None,
            "price": 0,
            "thumbnails": []}
    
    def pull_scene_data(self, scene: bpy.types.Scene, list_camera, EX_GLOBAL_MATRIX: Matrix):
        if scene.use_nodes:
            
            self.data["environment"] = {
                "background": {"r": 0.2, "g": 0.2, "b": 0.2, "a": 1},
                "ambient": {"r": 0.2, "g": 0.2, "b": 0.2, "a": 1},
                "skybox": {"material": "FXUnlitSkyboxLatlong"}
                }
        
        if len(list_camera) != 0:
            if scene.camera is not None:
                camera = scene.camera
            else:
                camera = list_camera[0]
            
            trl, rot, scl = EX_GLOBAL_MATRIX.decompose()
            px, py, pz = camera.location * scl
            quad = camera.rotation_quaternion.copy()
            quad.rotate(rot)
            
            self.data["camera"] = {
                "position": {"x": px, "y": pz, "z": -py},
                "orientation": {"w": quad[0], "x": quad[1], "y": quad[2], "z": quad[3]}}
